from .engine import XPlant, ModelNode


class PYamlNode(ModelNode):
    _single_indent = '  '

    def string_items(self, tree_level):
        representation = "{} - {}".format(self._single_indent * tree_level, self.own_value)
        if self.children:
            representation += ":"
        yield representation + "\n"

        for child in self.children:
            for item in child.string_items(tree_level + 1):
                yield item


class PYamlPlant(XPlant):
    regular_node_type = PYamlNode
    leaf_node_type = PYamlNode



from .engine import XPlant, ModelNode
from ._xml_attributes import XmlAttributes


class Text(ModelNode):
    def string_items(self, tree_level):
        yield self.own_value


class Comment(ModelNode):
    def string_items(self, tree_level):
        yield "<!-- %s -->" % self.own_value


class CdataTag(ModelNode):
    def string_items(self, tree_level):
        return "<![CDATA[%s]]>" % self.own_value


class _XmlElementBaseNode(ModelNode):
    attribute_processor = XmlAttributes
    _force_inline = False
    __slots__ = ("attributes",)

    def __init__(self, tag_name, *args, **kwargs):
        super(_XmlElementBaseNode, self).__init__(tag_name)
        self.attributes = self.attribute_processor(*args, **kwargs)


class EmptyXmlElement(_XmlElementBaseNode):
    """ E.g.: <br />, <hr /> or <img src="#" alt="" /> """

    def append_child_node(self, new_node_obj, plant):
        raise AttributeError("Cannot append to {} node.".format(self.__class__.__name__))

    def string_items(self, tree_level):
        yield "<{}{} />".format(self.own_value, self.attributes)


class XmlElement(_XmlElementBaseNode):
    _single_indent_val = "  "

    def string_items(self, tree_level):
        yield "<{}{}>".format(self.own_value, self.attributes)

        for item in self._childrens_markup(tree_level):
            yield item

        yield "</{}>".format(self.own_value)

    def _childrens_markup(self, tree_level):
        break_lines = self.degree > 1 and not self._force_inline
        child_indent = self._make_indent(tree_level + 1)

        for child in self.children:
            if break_lines:
                yield child_indent

            for child_string in child.string_items(tree_level + 1):
                yield child_string

        if break_lines:
            yield self._make_indent(tree_level)

    @classmethod
    def _make_indent(cls, tree_level):
        return "\n" + cls._single_indent_val * tree_level


class XmlPlant(XPlant):
    regular_node_type = XmlElement
    leaf_node_type = EmptyXmlElement

    def text(self, *text):
        for t in text:
            self.append_child_node(Text(t), self)

    def line(self, tag_name, content, *args, **kwargs):
        """ Element containing just a text. """
        with self.node(tag_name, *args, **kwargs):
            self.append_child_node(Text(content), self)

    def cdata(self, content):
        self.append_child_node(CdataTag(content), self)

    def comment(self, content):
        self.append_child_node(Comment(content), self)



from contextlib import contextmanager


class ModelNode(object):
    """
        Sequential node can describe itself in the document, yields markup items of itself and its children.
    """
    _empty_iterator = iter(())
    __slots__ = ("own_value", "children")

    def __init__(self, value):
        self.own_value = value
        self.children = []

    def __str__(self):
        return "".join(k for k in self.string_items(0) if k)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.own_value)

    def append_child_node(self, new_node_obj, plant):
        if not isinstance(new_node_obj, ModelNode):
            raise TypeError("Can append only instances of ModelNode, got {}.".format(type(new_node_obj).__name__))

        self.children.append(new_node_obj)

    @property
    def degree(self):
        """ For a given node, its number of children. A leaf is necessarily degree zero. """
        return len(self.children)

    @property
    def height(self):
        """ The height of a node is the number of edges on the longest path between that node and a leaf. """
        if not self.children:
            return 0
        else:
            return 1 + max(child.height for child in self.children)

    def string_items(self, tree_level):
        msg = "\n".join([
            "ModelNode.string_items is an abstract method.", "",
            "It's supposed to:",
            " - be a generator yielding string items,",
            " - compose document representation of that node with ''.join(node.string_items(0)),",
            " - care for document's line breaks, markup and indentation.",
            "It needs to be implemented in derived class {name!r}.",
            ""
        ])
        raise NotImplementedError(msg.format(name=self.__class__.__name__))


class RootNodeType(ModelNode):
    """
        Root node in current implementation has no owb value, no head nor tail markup.
        It's just a children's container.
    """

    def __init__(self):
        super(RootNodeType, self).__init__(None)

    def string_items(self, _):
        for child in self.children:
            for item in child.string_items(tree_level=0):
                yield item
        if self.children:
            yield "\n"


class LeafNodeType(ModelNode):
    """ Leaf - a node that is not allowed to have children. """

    def append_child_node(self, new_node_obj, plant):
        raise AttributeError("{} is not allowed to append any children".format(self.__class__.__name__))


class XPlant(object):
    _root_node_type = RootNodeType
    regular_node_type = ModelNode
    leaf_node_type = LeafNodeType

    def __init__(self):
        self.__tag_stack = [self._root_node_type()]

    def __str__(self):
        return str(self.__tag_stack[0])

    def __getattr__(self, attr_name):
        """ It exposes interface of current _top_node object. """
        if not attr_name.startswith("_") and hasattr(self._top_node, attr_name):
            return getattr(self._top_node, attr_name)
        return super(XPlant, self).__getattribute__(attr_name)

    @property
    def _top_node(self):
        """ The node whose context we are currently in. """
        return self.__tag_stack[-1]

    @contextmanager
    def node(self, *args, **kwargs):
        new_node = self.regular_node_type(*args, **kwargs)
        self._top_node.append_child_node(new_node, self)
        self.__tag_stack.append(new_node)
        try:
            yield
        finally:
            self.__tag_stack.pop()

    def leaf(self, *args, **kwargs):
        new_node = self.leaf_node_type(*args, **kwargs)
        self._top_node.append_child_node(new_node, self)

    @contextmanager
    def descent(self, new_node_obj):
        """ Register a new child node to the current self._top_node and enter its scope/context. """
        self._top_node.append_child_node(new_node_obj)

        self.__tag_stack.append(new_node_obj)
        try:
            yield
        finally:
            self.__tag_stack.pop()



from contextlib import contextmanager

from .xml import XmlPlant


class YattagPlant(XmlPlant):
    """ Exposes an interface similar to yattag's simpledoc. """

    @contextmanager
    def tag(self, tag_name, *args, **kwargs):
        with self.node(tag_name, *args, **kwargs):
            yield

    def stag(self, tag_name, *args, **kwargs):
        """ Empty xml element. """
        self.leaf(tag_name, *args, **kwargs)

    def tagtext(self):
        return self, self.tag, self.text

    def ttl(self):
        return self, self.tag, self.text, self.line

    def getvalue(self):
        return str(self)

    def cdata(self, content, safe=False):
        if safe:
            content = content.replace(']]>', ']]]]><![CDATA[>')
        super(YattagPlant, self).cdata(content)
